""" Exmples of usage with tests.

Tests in this file represent examples taken from JSON-RPC specification.
http://www.jsonrpc.org/specification#examples

"""
import sys
if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest
import json
from ..manager import JSONRPCResponseManager


def isjsonequal(json1, json2):
    return json.loads(json1) == json.loads(json2)


class TestJSONRPCExamples(unittest.TestCase):
    def setUp(self):
        self.dispatcher = {
            "return_none": lambda: None,
        }

    def test_none_as_result(self):
        req = '{"jsonrpc": "2.0", "method": "return_none", "id": 0}'
        response = JSONRPCResponseManager.handle(req, self.dispatcher)
        self.assertTrue(isjsonequal(
            response.json,
            '{"jsonrpc": "2.0", "result": null, "id": 0}'
        ))




# coding=utf-8


class IConnection(object):

    def send(self, msg):
        raise NotImplementedError

    def recv(self):
        raise NotImplementedError


class IClient(IConnection):
    def __init__(self):
        super(IClient, self).__init__()
        self.connect_cb = None  # callable()
        self.close_cb = None  # callable()

    def on_connect(self):
        """
        this function must be called on_connect
        """
        if callable(self.connect_cb):
            self.connect_cb()

    def on_close(self):
        """
        this function must be called on_close
        """
        if callable(self.close_cb):
            self.close_cb()

    def connect(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError


class IServer(object):

    def __init__(self):
        super(IServer, self).__init__()
        self.client_connect_cb = None  # callable(conn)
        self.client_close_cb = None  # callable(conn)

    def on_client_connect(self, conn):
        """
        this function must be called on_client_connect
        """
        if callable(self.client_connect_cb):
            self.client_connect_cb(conn)

    def on_client_close(self, conn):
        """
        this function must be called on_client_close
        """
        if callable(self.client_close_cb):
            self.client_close_cb(conn)

    @property
    def connections(self):
        # return {cid: IConnection}
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def broadcast(self, msg):
        for conn in self.connections.values():
            conn.send(msg)



# encoding=utf-8
from .simplerpc import RpcAgent


class RpcServer(RpcAgent):
    """docstring for RpcServer"""
    def __init__(self, server):
        super(RpcServer, self).__init__()
        self.server = server
        self.server.client_connect_cb = self.on_client_connect
        self.server.client_close_cb = self.on_client_close
        self.server.start()

    def on_client_connect(self, conn):
        print("on_client_connect", conn)

    def on_client_close(self, conn):
        print("on_client_close", conn)

    def update(self):
        for conn in self.server.connections.values():
            messages = conn.recv()
            for msg in messages:
                self.handle_message(msg, conn)

    def call(self, conn, func, *args, **kwargs):
        req, cb = self.format_request(func, *args, **kwargs)
        conn.send(req)
        return cb



# -*- coding: utf-8 -*-
# @Author: gzliuxin
# @Email:  gzliuxin@corp.netease.com
# @Date:   2017-07-14 19:47:51

from poco.drivers.std import StdPoco

from airtest.core.api import connect_device, device as current_device


__all__ = ['UnityPoco']
DEFAULT_PORT = 5001
DEFAULT_ADDR = ("localhost", DEFAULT_PORT)


class UnityPoco(StdPoco):
    """
    Poco Unity3D implementation.

    Args:
        addr (:py:obj:`tuple`): the endpoint of your Unity3D game, default to ``("localhost", 5001)``
        unity_editor (:py:obj:`bool`): whether your Unity3D game is running in UnityEditor or not. default to ``False``
        connect_default_device (:py:obj:`bool`): whether connect to a default device if no devices selected manually.
         default to ``True``.
        options: see :py:class:`poco.pocofw.Poco`

    Examples:
        If your game is running on Android, you could initialize poco instance by using following snippet::
            
            from poco.drivers.unity3d import UnityPoco
            
            # your phone and your PC/mac should be inside the same sub-net.
            ip = '<ip address of your phone>'
            poco = UnityPoco((ip, 5001))
            poco('button').click()
            ...

    """

    def __init__(self, addr=DEFAULT_ADDR, unity_editor=False, connect_default_device=True, **options):
        if 'action_interval' not in options:
            options['action_interval'] = 0.5

        dev = None
        if unity_editor:
            dev = connect_device("Windows:///?class_name=UnityWndClass&title_re=Unity.*")
            game_window = dev.app.top_window().child_window(title="UnityEditor.GameView")
            dev._top_window = game_window.wrapper_object()
            dev.focus_rect = (0, 40, 0, 0)

        if connect_default_device and not current_device():
            # currently only connect to Android as default
            # can apply auto detection in the future
            dev = connect_device("Android:///")

        super(UnityPoco, self).__init__(addr[1], dev, **options)



# coding=utf-8
from .DefaultMatcher import DefaultMatcher
from .exceptions import NoSuchTargetException

__author__ = 'lxn3032'
__all__ = ['ISelector', 'Selector']


class ISelector(object):
    """
    This interface defines the standard selector behavior. Selector is used for selecting the specific UI element(s)
    by given query expression (formal definitions are in specific implementation classes).
    """

    def select(self, cond, multiple=False):
        """
        Args:
            cond (:obj:`tuple`): query expressiom
            multiple (:obj:`bool`): whether or not to select the multiple elements. If set to True, the method
                                    terminates immediately once the node is found, otherwise it traverses through all
                                    nodes and then exists

        Returns:
            :obj:`list`: list <inherited from :py:class:`AbstractNode <poco.sdk.AbstractNode>`>
        """

        raise NotImplementedError


class Selector(ISelector):
    """
    This class implements the standard Selector interface that uses BFS algorithm to travers through tree-like hierarchy
    structure. It supports flexible query expressions such as parental relationship, attribute predicate, etc. Any
    combinations of expressions mentioned above are also allowed as the query conditions.
    
    The query expression can be defined as follows::

        expr := (op0, (expr0, expr1))
        expr := ('index', (expr, :obj:`int`))
        expr := <other query condition> See implementation of Matcher.

    - ``op0`` can be one of the following ('>', '/', '-'), each operator stands for as follows::

        '>': offsprings, select all offsprings matched expr1 from all roots matched expr0.
        '/': children, select all children matched expr1 from all roots matched expr0.
        '-': siblings, select all siblings matched expr1 from all roots matched expr0.
    
    - ``'index'``: select specific n-th UI element from the previous results

    - ``others``: passes the expression to matcher

    Args:
        dumper (any implementation of :py:class:`IDumper <poco.sdk.AbstractDumper.IDumper>`):  dumper for the selector
        matcher (any implementation of :py:class:`IMatcher <poco.sdk.DefaultMatcher.IMatcher>`): :py:class:`DefaultMatcher
         <poco.sdk.DefaultMatcher.DefaultMatcher>` instance by default.
    """

    def __init__(self, dumper, matcher=None):
        self.dumper = dumper
        self.matcher = matcher or DefaultMatcher()

    def getRoot(self):
        """
        Get a default root node.

        Returns:
            default root node from the dumper.
        """

        return self.dumper.getRoot()

    def select(self, cond, multiple=False):
        """
        See Also: :py:meth:`select <poco.sdk.Selector.ISelector.select>` method in ``ISelector``.
        """

        return self.selectImpl(cond, multiple, self.getRoot(), 9999, True, True)

    def selectImpl(self, cond, multiple, root, maxDepth, onlyVisibleNode, includeRoot):
        """
        Selector internal implementation. 
        TODO: add later.

        .. note:: This doc shows only the outline of the algorithm. Do not call this method in your code as this
         is an internal method.

        Args:
            cond (:obj:`tuple`): query expression
            multiple (:obj:`bool`): whether or not to select multiple nodes. If true, all nodes that matches the given
             condition will return, otherwise, only the first node matches will.
            root (inherited from :py:class:`AbstractNode <poco.sdk.AbstractNode>`): start traversing from the given
             root node
            maxDepth (:obj:`bool`): max traversing depth
            onlyVisibleNode (:obj:`bool`): If True, skip those node which visibility (the value of visible attribute) is
             False.
            includeRoot (:obj:`bool`): whether not not to include the root node if its child(ren) match(es) the node

        Returns:
            :obj:`list` <inherited from :py:class:`AbstractNode <poco.sdk.AbstractNode>`>: The same as
             :py:meth:`select <poco.sdk.Selector.ISelector.select>`.
        """

        result = []
        if not root:
            return result

        op, args = cond

        if op in ('>', '/'):
            # children or offsprings
            # 父子直系相对节点选择
            parents = [root]
            for index, arg in enumerate(args):
                midResult = []
                for parent in parents:
                    if op == '/' and index != 0:
                        _maxDepth = 1
                    else:
                        _maxDepth = maxDepth
                    # 按路径进行遍历一定要multiple为true才不会漏掉
                    _res = self.selectImpl(arg, True, parent, _maxDepth, onlyVisibleNode, False)
                    [midResult.append(r) for r in _res if r not in midResult]
                parents = midResult
            result = parents
        elif op == '-':
            # sibling
            # 兄弟节点选择
            query1, query2 = args
            result1 = self.selectImpl(query1, multiple, root, maxDepth, onlyVisibleNode, includeRoot)
            for n in result1:
                sibling_result = self.selectImpl(query2, multiple, n.getParent(), 1, onlyVisibleNode, includeRoot)
                [result.append(r) for r in sibling_result if r not in result]
        elif op == 'index':
            cond, i = args
            try:
                result = [self.selectImpl(cond, multiple, root, maxDepth, onlyVisibleNode, includeRoot)[i]]
            except IndexError:
                raise NoSuchTargetException(
                    u'Query results index out of range. Index={} condition "{}" from root "{}".'.format(i, cond, root))
        else:
            self._selectTraverse(cond, root, result, multiple, maxDepth, onlyVisibleNode, includeRoot)

        return result

    def _selectTraverse(self, cond, node, outResult, multiple, maxDepth, onlyVisibleNode, includeRoot):
        # exclude invisible UI element if onlyVisibleNode specified
        # 剪掉不可见节点branch
        if onlyVisibleNode and not node.getAttr('visible'):
            return False

        if self.matcher.match(cond, node):
            # To select node from parent or ancestor, the parent or ancestor are excluded.
            # 父子/祖先后代节点选择时，默认是不包含父节点/祖先节点的
            # 在下面的children循环中则需要包含，因为每个child在_selectTraverse中就当做是root
            if includeRoot:
                if node not in outResult:
                    outResult.append(node)
                if not multiple:
                    return True

        # When maximum search depth reached, children of this node is still require to travers.
        # 最大搜索深度耗尽并不表示遍历结束，其余child节点仍需遍历
        if maxDepth == 0:
            return False
        maxDepth -= 1

        for child in node.getChildren():
            finished = self._selectTraverse(cond, child, outResult, multiple, maxDepth, onlyVisibleNode, True)
            if finished:
                return True

        return False



""" Dispatcher is used to add methods (functions) to the server.

For usage examples see :meth:`Dispatcher.add_method`

"""
import collections


class Dispatcher(collections.MutableMapping):

    """ Dictionary like object which maps method_name to method."""

    def __init__(self, prototype=None):
        """ Build method dispatcher.

        Parameters
        ----------
        prototype : object or dict, optional
            Initial method mapping.

        Examples
        --------

        Init object with method dictionary.

        >>> Dispatcher({"sum": lambda a, b: a + b})
        None

        """
        self.method_map = dict()

        if prototype is not None:
            self.build_method_map(prototype)

    def __getitem__(self, key):
        return self.method_map[key]

    def __setitem__(self, key, value):
        self.method_map[key] = value

    def __delitem__(self, key):
        del self.method_map[key]

    def __len__(self):
        return len(self.method_map)

    def __iter__(self):
        return iter(self.method_map)

    def __repr__(self):
        return repr(self.method_map)

    def add_class(self, cls):
        prefix = cls.__name__.lower() + '.'
        self.build_method_map(cls(), prefix)

    def add_object(self, obj):
        prefix = obj.__class__.__name__.lower() + '.'
        self.build_method_map(obj, prefix)

    def add_dict(self, dict, prefix=''):
        if prefix:
            prefix += '.'
        self.build_method_map(dict, prefix)

    def add_method(self, f, name=None):
        """ Add a method to the dispatcher.

        Parameters
        ----------
        f : callable
            Callable to be added.
        name : str, optional
            Name to register (the default is function **f** name)

        Notes
        -----
        When used as a decorator keeps callable object unmodified.

        Examples
        --------

        Use as method

        >>> d = Dispatcher()
        >>> d.add_method(lambda a, b: a + b, name="sum")
        <function __main__.<lambda>>

        Or use as decorator

        >>> d = Dispatcher()
        >>> @d.add_method
            def mymethod(*args, **kwargs):
                print(args, kwargs)

        """
        self.method_map[name or f.__name__] = f
        return f

    def build_method_map(self, prototype, prefix=''):
        """ Add prototype methods to the dispatcher.

        Parameters
        ----------
        prototype : object or dict
            Initial method mapping.
            If given prototype is a dictionary then all callable objects will
            be added to dispatcher.
            If given prototype is an object then all public methods will
            be used.
        prefix: string, optional
            Prefix of methods

        """
        if not isinstance(prototype, dict):
            prototype = dict((method, getattr(prototype, method))
                             for method in dir(prototype)
                             if not method.startswith('_'))

        for attr, method in prototype.items():
            if callable(method):
                self[prefix + attr] = method



# coding=utf-8

from poco.sdk.AbstractDumper import AbstractDumper
from poco.sdk.AbstractNode import AbstractNode
from poco.sdk.Attributor import Attributor
from poco.sdk.Selector import Selector
from poco.sdk.exceptions import UnableToSetAttributeException
from poco.sdk.interfaces.hierarchy import HierarchyInterface


__all__ = ['FrozenUIDumper', 'FrozenUIHierarchy']


class FrozenUIDumper(AbstractDumper):
    """
    Partially implementation of IDumper. This is only a helper that helps to make dumper work with local nodes just 
    like with remote nodes. The local nodes is an implementation of :py:class:`AbstractNode <poco.sdk.AbstractNode>` 
    locally with fixed hierarchy data in arbitrary data type. In a word, this class is not going to crawl hierarchy from 
    target app, but to perform like a ordinary dumper.
    """

    def dumpHierarchy(self, onlyVisibleNode=True):
        raise NotImplementedError

    def getRoot(self):
        """
        Dump a hierarchy immediately from target runtime and store into a Node (subclass of :py:class:`AbstractNode 
        <poco.sdk.AbstractNode>`) object.

        Returns:
            :py:class:`inherit from AbstractNode <Node>`: Each time a new node instance is created by latest hierarchy 
             data.
        """

        root = Node(self.dumpHierarchy())
        self._linkParent(root)
        return root

    def _linkParent(self, root):
        parent = root.getChildren()
        if parent:
            for child in parent:
                child.setParent(root)
                self._linkParent(child)


class FrozenUIHierarchy(HierarchyInterface):
    """
    Locally implementation of hierarchy interface with a given dumper and all other behaviours by default. As all 
    information can only be retrieve from a fixed UI hierarchy data created by dumper and all UI elements are immutable, 
    this class is called frozen hierarchy. With the help of this class, only very few of the methods are required to
    implement. See :py:class:`AbstractNode <poco.sdk.AbstractNode>` or poco-sdk integration guide to get more details 
    about this. 
     
    This class makes it much easier to integrate poco-sdk and optimizes performance in some situations, but it is not 
    sensitive enough to the changes of UI hierarchy in the app. For example, you should call ``select`` explicitly to 
    re-crawl a new UI hierarchy when some UI elements changed on screen. Otherwise you are using attributes that are 
    out of date.
    """

    def __init__(self, dumper, attributor=None):
        super(FrozenUIHierarchy, self).__init__()
        self.dumper = dumper
        self.selector = Selector(self.dumper)
        self.attributor = attributor or Attributor()

    def dump(self):
        return self.dumper.dumpHierarchy()

    def getAttr(self, nodes, name):
        """
        get node attribute
        """

        return self.attributor.getAttr(nodes, name)

    def setAttr(self, nodes, name, value):
        """
        set node attribute
        """

        return self.attributor.setAttr(nodes, name, value)

    def select(self, query, multiple=False):
        """
        select nodes by query
        """

        return self.selector.select(query, multiple)


class Node(AbstractNode):
    def __init__(self, node):
        super(Node, self).__init__()
        self.node = node

    def setParent(self, p):
        self.node['__parent__'] = p

    def getParent(self):
        return self.node.get('__parent__')

    def getChildren(self):
        for child in self.node.get('children') or []:
            yield Node(child)

    def getAttr(self, attrName):
        return self.node['payload'].get(attrName)

    def setAttr(self, attrName, val):
        # cannot set any attributes on local nodes
        raise UnableToSetAttributeException(attrName, self.node)

    def getAvailableAttributeNames(self):
        return self.node['payload'].keys()



import json
import sys
from mock import patch

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

# Flask is supported only for python2 and pyton3.3+
if sys.version_info < (3, 0) or sys.version_info >= (3, 3):
    from flask import Flask
    from ...backend.flask import JSONRPCAPI, api

    @api.dispatcher.add_method
    def dummy():
        return ""


@unittest.skipIf((3, 0) <= sys.version_info < (3, 3),
                 'Flask does not support python 3.0 - 3.2')
class TestFlaskBackend(unittest.TestCase):
    REQUEST = json.dumps({
        "id": "0",
        "jsonrpc": "2.0",
        "method": "dummy",
    })

    def setUp(self):
        self.client = self._get_test_client(JSONRPCAPI())

    def _get_test_client(self, api):
        @api.dispatcher.add_method
        def dummy():
            return ""

        app = Flask(__name__)
        app.config["TESTING"] = True
        app.register_blueprint(api.as_blueprint())
        return app.test_client()

    def test_client(self):
        response = self.client.post(
            '/',
            data=self.REQUEST,
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['result'], '')

    def test_method_not_allowed(self):
        response = self.client.get(
            '/',
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 405, "Should allow only POST")

    def test_parse_error(self):
        response = self.client.post(
            '/',
            data='{',
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['error']['code'], -32700)
        self.assertEqual(data['error']['message'], 'Parse error')

    def test_wrong_content_type(self):
        response = self.client.post(
            '/',
            data=self.REQUEST,
            content_type='application/x-www-form-urlencoded',
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['error']['code'], -32700)
        self.assertEqual(data['error']['message'], 'Parse error')

    def test_invalid_request(self):
        response = self.client.post(
            '/',
            data='{"method": "dummy", "id": 1}',
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['error']['code'], -32600)
        self.assertEqual(data['error']['message'], 'Invalid Request')

    def test_method_not_found(self):
        data = {
            "jsonrpc": "2.0",
            "method": "dummy2",
            "id": 1
        }
        response = self.client.post(
            '/',
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['error']['code'], -32601)
        self.assertEqual(data['error']['message'], 'Method not found')

    def test_invalid_parameters(self):
        data = {
            "jsonrpc": "2.0",
            "method": "dummy",
            "params": [42],
            "id": 1
        }
        response = self.client.post(
            '/',
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['error']['code'], -32602)
        self.assertEqual(data['error']['message'], 'Invalid params')

    def test_resource_map(self):
        response = self.client.get('/map')
        self.assertEqual(response.status_code, 200)
        self.assertTrue("JSON-RPC map" in response.data.decode('utf8'))

    def test_method_not_allowed_prefix(self):
        response = self.client.get(
            '/',
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 405)

    def test_resource_map_prefix(self):
        response = self.client.get('/map')
        self.assertEqual(response.status_code, 200)

    def test_as_view(self):
        api = JSONRPCAPI()
        with patch.object(api, 'jsonrpc') as mock_jsonrpc:
            self.assertIs(api.as_view(), mock_jsonrpc)

    def test_not_check_content_type(self):
        client = self._get_test_client(JSONRPCAPI(check_content_type=False))
        response = client.post(
            '/',
            data=self.REQUEST,
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['result'], '')

    def test_check_content_type(self):
        client = self._get_test_client(JSONRPCAPI(check_content_type=False))
        response = client.post(
            '/',
            data=self.REQUEST,
            content_type="application/x-www-form-urlencoded"
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data.decode('utf8'))
        self.assertEqual(data['result'], '')

    def test_empty_initial_dispatcher(self):
        class SubDispatcher(type(api.dispatcher)):
            pass

        custom_dispatcher = SubDispatcher()
        custom_api = JSONRPCAPI(custom_dispatcher)
        self.assertEqual(type(custom_api.dispatcher), SubDispatcher)
        self.assertEqual(id(custom_api.dispatcher), id(custom_dispatcher))



# -*- coding: utf-8 -*-
# @Author: gzliuxin
# @Email:  gzliuxin@corp.netease.com
# @Date:   2017-07-12 16:56:14
import json
import time
import traceback

from .jsonrpc import JSONRPCResponseManager, dispatcher
from .jsonrpc.jsonrpc2 import JSONRPC20Response
from .jsonrpc.exceptions import JSONRPCServerError
from .jsonrpc import six


DEBUG = True
BACKEND_UPDATE = False


class Callback(object):
    """Callback Proxy"""

    WAITING, RESULT, ERROR, CANCELED = 0, 1, 2, 3

    def __init__(self, rid, agent=None):
        super(Callback, self).__init__()
        self.rid = rid
        self.agent = agent
        self.result_callback = None
        self.error_callback = None
        self.status = self.WAITING
        self.result = None
        self.error = None

    def on_result(self, func):
        if not callable(func):
            raise RuntimeError("%s should be callbale" % func)
        self.result_callback = func

    def on_error(self, func):
        if not callable(func):
            raise RuntimeError("%s should be callbale" % func)
        self.error_callback = func

    def rpc_result(self, data):
        self.result = data
        if callable(self.result_callback):
            # callback function, set result as function return value
            try:
                self.result_callback(data)
            except Exception:
                traceback.print_exc()
        self.status = self.RESULT

    def rpc_error(self, data):
        self.error = data
        if callable(self.error_callback):
            try:
                self.error_callback(data)
            except Exception:
                traceback.print_exc()
        self.status = self.ERROR

    def cancel(self):
        self.result_callback = None
        self.error_callback = None
        self.status = self.CANCELED

    def wait(self, timeout=None):
        start_time = time.time()
        while True:
            if not BACKEND_UPDATE:
                self.agent.update()
            if self.status == self.WAITING:
                time.sleep(0.005)
                if timeout and time.time() - start_time > timeout:
                    raise RuntimeError("%s timeout error" % self)
            else:
                break

        return (self.result, self.error)


class AsyncResponse(object):

    def __init__(self):
        self.conn = None
        self.rid = None

    def setup(self, conn, rid):
        self.conn = conn
        self.rid = rid

    def result(self, result):
        ret = JSONRPC20Response(_id=self.rid, result=result)
        if DEBUG:
            print("-->", ret)

        self.conn.send(ret.json)

    def error(self, error):
        assert isinstance(error, Exception), "%s must be Exception" % error
        data = {
            "type": error.__class__.__name__,
            "args": error.args,
            "message": str(error),
        }
        ret = JSONRPC20Response( _id=self.rid, error=JSONRPCServerError(data=data)._data)
        if DEBUG:
            print("-->", ret)
        self.conn.send(ret.json)


class RpcAgent(object):
    """docstring for RpcAgent"""

    REQUEST = 0
    RESPONSE = 1

    def __init__(self):
        super(RpcAgent, self).__init__()
        self._id = 0
        self._callbacks = {}

    def call(self, *args, **kwargs):
        raise NotImplementedError

    def format_request(self, func, *args, **kwargs):
        rid = self._id
        payload = {
            "method": func,
            "params": args or kwargs or [],
            "jsonrpc": "2.0",
            "id": rid,
        }
        self._id += 1
        # send rpc
        req = json.dumps(payload)
        if DEBUG:
            print("-->", req)
        # init cb
        cb = Callback(rid, self)
        self._callbacks[rid] = cb
        return req, cb

    def handle_request(self, req):
        res = JSONRPCResponseManager.handle(req, dispatcher).data
        return res

    def handle_message(self, msg, conn):
        if isinstance(msg, six.binary_type):
            # py3里 json 只接受str类型，py2没有这个限制
            msg = msg.decode('utf-8')
        data = json.loads(msg)
        # if DEBUG:
        #     print("<--", data)
        if "method" in data:
            # rpc request
            message_type = self.REQUEST
            result = self.handle_request(msg)

            if isinstance(result.get("result"), AsyncResponse):
                result["result"].setup(conn, result["id"])
            else:
                # if DEBUG:
                #     print("-->", result)
                conn.send(json.dumps(result))

        else:
            # rpc response
            message_type = self.RESPONSE
            result = None
            # handle callback
            callback = self._callbacks.pop(data["id"])
            if "result" in data:
                callback.rpc_result(data["result"])
            elif "error" in data:
                callback.rpc_error(data["error"])
            else:
                pass
        return message_type, result

    def update(self):
        raise NotImplementedError

    def run(self):
        def _run():
            while True:
                self.update()
                time.sleep(0.002)
        if BACKEND_UPDATE:
            from threading import Thread
            t = Thread(target=_run, name="update")
            t.daemon = True
            t.start()
        else:
            _run()

    def console_run(self, local_dict=None):
        global BACKEND_UPDATE
        BACKEND_UPDATE = True
        self.run()
        from code import InteractiveInterpreter
        i = InteractiveInterpreter(local_dict)
        while True:
            prompt = ">>>"
            try:
                line = raw_input(prompt)
            except EOFError:
                print("closing..")
                return
            i.runcode(line)



# coding=utf-8

try:
    from airtest.core.api import device as current_device
    from airtest.core.api import touch, swipe
    from airtest.core.helper import device_platform
except ImportError:
    # 兼容旧版本
    from airtest.cli.runner import device as current_device
    from airtest.core.main import touch, swipe
    from airtest.core.main import get_platform as device_platform
from poco.sdk.interfaces.input import InputInterface
from poco.utils.track import track_sampling


class AirtestInput(InputInterface):
    def __init__(self):
        super(AirtestInput, self).__init__()
        self.default_touch_down_duration = 0.01

    def _get_touch_resolution(self):
        """
        get real time resolution on device if full screen
         or window size if running in window mode
        """
        return current_device().get_current_resolution()

    def setTouchDownDuration(self, duration):
        self.default_touch_down_duration = duration

    def getTouchDownDuration(self):
        return self.default_touch_down_duration

    def click(self, x, y):
        pw, ph = self._get_touch_resolution()
        pos = [x * pw, y * ph]
        touch(pos, duration=self.default_touch_down_duration)

    def swipe(self, x1, y1, x2, y2, duration=2.0):
        if duration <= 0:
            raise ValueError("Operation duration cannot be less equal 0. Please provide a positive number.")
        direction = x2 - x1, y2 - y1
        pw, ph = self._get_touch_resolution()
        p1 = [x1 * pw, y1 * ph]
        steps = int(duration * 40) + 1
        swipe(p1, vector=direction, duration=duration, steps=steps)

    def longClick(self, x, y, duration=2.0):
        if duration <= 0:
            raise ValueError("Operation duration cannot be less equal 0. Please provide a positive number.")
        pw, ph = self._get_touch_resolution()
        pos = [x * pw, y * ph]
        touch(pos, duration=duration)

    def applyMotionEvents(self, events):
        if device_platform() != 'Android':
            raise NotImplementedError

        # Android minitouch only, currently
        from airtest.core.android.minitouch import DownEvent, MoveEvent, UpEvent, SleepEvent

        w, h = self._get_touch_resolution()
        mes = []
        for e in events:
            t = e[0]
            if t == 'd':
                contact = e[2]
                x, y = e[1]
                me = DownEvent([x * w, y * h], contact)
            elif t == 'm':
                contact = e[2]
                x, y = e[1]
                me = MoveEvent([x * w, y * h], contact)
            elif t == 'u':
                contact = e[1]
                me = UpEvent(contact)
            elif t == 's':
                how_long = e[1]
                me = SleepEvent(how_long)
            else:
                raise ValueError('Unknown event type {}'.format(repr(t)))
            mes.append(me)

        current_device().minitouch.perform(mes, interval=0)



# coding=utf-8

import time
from poco.exceptions import InvalidOperationException
from poco.drivers.unity3d.test.tutorial.case import TutorialCase


class InvalidOperationExceptionTutorial(TutorialCase):
    def runTest(self):
        try:
            self.poco.click([1.1, 1.1])  # click outside screen
        except InvalidOperationException:
            print('oops')
            time.sleep(1)


if __name__ == '__main__':
    import pocounit
    pocounit.main()



# coding=utf-8
from functools import wraps


class RemoteError(Exception):
    pass


def sync_wrapper(func):
    @wraps(func)
    def new_func(*args, **kwargs):
        cb = func(*args, **kwargs)
        ret, err = cb.wait(timeout=30)
        if err:
            raise RemoteError(err['message'])
        return ret
    return new_func



# coding=utf-8

from typing import Iterable, Iterator, Text, Tuple, Union, List, Dict, Any

from poco.sdk.interfaces.hierarchy import HierarchyInterface
from poco.sdk.interfaces.input import InputInterface
from poco.sdk.interfaces.screen import ScreenInterface
from poco.sdk.interfaces.command import CommandInterface


class PocoAgent(object):
    def __init__(self,
                 hierarchy: HierarchyInterface,
                 input: InputInterface,
                 screen: ScreenInterface,
                 command: CommandInterface=None):
        self.hierarchy = ...    # type: HierarchyInterface
        self.input = ...        # type: InputInterface
        self.screen = ...       # type: ScreenInterface
        self.command = ...      # type: CommandInterface

    def get_sdk_version(self) -> Text:
        ...



# coding=utf-8

from poco.freezeui.hierarchy import FrozenUIDumper, FrozenUIHierarchy


__all__ = ['create_immutable_hierarchy', 'create_immutable_dumper']


def create_immutable_hierarchy(hierarchy_dict):
    dumper = create_immutable_dumper(hierarchy_dict)
    return FrozenUIHierarchy(dumper)


def create_immutable_dumper(hierarchy_dict):
    class ImmutableFrozenUIDumper(FrozenUIDumper):
        def dumpHierarchy(self, onlyVisibleNode=True):
            return hierarchy_dict

    return ImmutableFrozenUIDumper()



# coding=utf-8
__author__ = 'lxn3032'


import os
import requests
import time
import warnings
import atexit

try:
    new_airtest_api = True
    import airtest.core.api as _____
except ImportError:
    new_airtest_api = False
    from airtest.core.android.utils.iputils import get_ip_address

from airtest.core.android.ime import YosemiteIme

from hrpc.client import RpcClient
from hrpc.transport.http import HttpTransport
from poco import Poco
from poco.agent import PocoAgent
from poco.sdk.Attributor import Attributor
from poco.sdk.interfaces.screen import ScreenInterface
from poco.utils.hrpc.hierarchy import RemotePocoHierarchy
from poco.utils.airtest.input import AirtestInput
from poco.utils import six
from poco.drivers.android.utils.installation import install, uninstall

__all__ = ['AndroidUiautomationPoco', 'AndroidUiautomationHelper']
this_dir = os.path.dirname(os.path.realpath(__file__))
PocoServicePackage = 'com.netease.open.pocoservice'
PocoServicePackageTest = 'com.netease.open.pocoservice.test'


class AndroidRpcClient(RpcClient):
    def __init__(self, endpoint):
        self.endpoint = endpoint
        super(AndroidRpcClient, self).__init__(HttpTransport)

    def initialize_transport(self):
        return HttpTransport(self.endpoint, self)


class AttributorWrapper(Attributor):
    """
    部分手机上仍不支持Accessibility.ACTION_SET_TEXT，使用YosemiteIme还是兼容性最好的方案
    这个class会hook住set_text，然后改用ime的text方法
    """

    def __init__(self, remote, ime):
        self.remote = remote
        self.ime = ime

    def getAttr(self, node, attrName):
        return self.remote.getAttr(node, attrName)

    def setAttr(self, node, attrName, attrVal):
        if attrName == 'text' and attrVal != '':
            # 先清除了再设置，虽然这样不如直接用ime的方法好，但是也能凑合用着
            current_val = self.remote.getAttr(node, 'text')
            if current_val:
                self.remote.setAttr(node, 'text', '')
            self.ime.text(attrVal)
        else:
            self.remote.setAttr(node, attrName, attrVal)


class ScreenWrapper(ScreenInterface):
    def __init__(self, screen):
        super(ScreenWrapper, self).__init__()
        self.screen = screen

    def getScreen(self, width):
        # Android上PocoService的实现为仅返回b64编码的图像，格式固定位jpg
        b64img = self.screen.getScreen(width)
        return b64img, 'jpg'

    def getPortSize(self):
        return self.screen.getPortSize()


class AndroidPocoAgent(PocoAgent):
    def __init__(self, endpoint, ime, use_airtest_input=False):
        self.client = AndroidRpcClient(endpoint)
        remote_poco = self.client.remote('poco-uiautomation-framework')
        dumper = remote_poco.dumper
        selector = remote_poco.selector
        attributor = AttributorWrapper(remote_poco.attributor, ime)
        hierarchy = RemotePocoHierarchy(dumper, selector, attributor)

        if use_airtest_input:
            inputer = AirtestInput()
        else:
            inputer = remote_poco.inputer
        super(AndroidPocoAgent, self).__init__(hierarchy, inputer, ScreenWrapper(remote_poco.screen), None)


class AndroidUiautomationPoco(Poco):
    """
    Poco Android implementation for testing **Android native apps**.

    Args:
        device (:py:obj:`Device`): :py:obj:`airtest.core.device.Device` instance provided by ``airtest``. leave the 
         parameter default and the default device will be chosen. more details refer to ``airtest doc``
        using_proxy (:py:obj:`bool`): whether use adb forward to connect the Android device or not
        force_restart (:py:obj:`bool`): whether always restart the poco-service-demo running on Android device or not
        options: see :py:class:`poco.pocofw.Poco`

    Examples:
        The simplest way to initialize AndroidUiautomationPoco instance and no matter your device network status::

            from poco.drivers.android.uiautomation import AndroidUiautomationPoco

            poco = AndroidUiautomationPoco()
            poco('android:id/title').click()
            ...

    """

    def __init__(self, device=None, using_proxy=True, force_restart=False, use_airtest_input=False, **options):
        # 加这个参数为了不在最新的pocounit方案中每步都截图
        self.screenshot_each_action = True
        if options.get('screenshot_each_action') is False:
            self.screenshot_each_action = False

        if not device:
            try:
                # new version
                from airtest.core.api import connect_device, device as current_device
                if not current_device():
                    connect_device("Android:///")
            except ImportError:
                # old version
                from airtest.cli.runner import device as current_device
                from airtest.core.main import set_serialno
                if not current_device():
                    set_serialno()
            self.device = current_device()
        else:
            self.device = device
        self.adb_client = self.device.adb
        if using_proxy:
            self.device_ip = self.adb_client.host or "127.0.0.1"
        else:
            if new_airtest_api:
                self.device_ip = self.adb_client.get_ip_address()
            else:
                self.device_ip = get_ip_address(self.adb_client)

        # save current top activity (@nullable)
        current_top_activity_package = self.device.get_top_activity_name()
        if current_top_activity_package is not None:
            current_top_activity_package = current_top_activity_package.split('/')[0]

        # install ime
        if new_airtest_api:
            self.ime = YosemiteIme(self.adb_client)
        else:
            self.ime = YosemiteIme(self.device)
        self.ime.start()

        # install
        self._instrument_proc = None
        self._install_service()

        # forward
        if using_proxy:
            p0, _ = self.adb_client.setup_forward("tcp:10080")
            p1, _ = self.adb_client.setup_forward("tcp:10081")
        else:
            p0 = 10080
            p1 = 10081

        # start
        if self._is_running('com.github.uiautomator'):
            warnings.warn('{} should not run together with "uiautomator". "uiautomator" will be killed.'
                          .format(self.__class__.__name__))
            self.adb_client.shell(['am', 'force-stop', 'com.github.uiautomator'])

        ready = self._start_instrument(p0, force_restart=force_restart)
        if not ready:
            # 启动失败则需要卸载再重启，instrument的奇怪之处
            uninstall(self.adb_client, PocoServicePackage)
            self._install_service()
            ready = self._start_instrument(p0)

            if current_top_activity_package is not None:
                current_top_activity2 = self.device.get_top_activity_name()
                if current_top_activity2 is None or current_top_activity_package not in current_top_activity2:
                    self.device.start_app(current_top_activity_package, activity=True)

            if not ready:
                raise RuntimeError("unable to launch AndroidUiautomationPoco")

        endpoint = "http://{}:{}".format(self.device_ip, p1)
        agent = AndroidPocoAgent(endpoint, self.ime, use_airtest_input)
        super(AndroidUiautomationPoco, self).__init__(agent, **options)

    def _install_service(self):
        updated = install(self.adb_client, os.path.join(this_dir, 'lib', 'pocoservice-debug.apk'))
        install(self.adb_client, os.path.join(this_dir, 'lib', 'pocoservice-debug-androidTest.apk'), updated)
        return updated

    def _is_running(self, package_name):
        processes = self.adb_client.shell(['ps']).splitlines()
        for ps in processes:
            ps = ps.strip()
            if ps.endswith(package_name):
                return True
        return False

    # def _keep_running_instrumentation(self):
    #     def loop():
    #         while True:
    #             proc = self.adb_client.shell([
    #                 'am', 'instrument', '-w', '-e', 'class',
    #                 '{}.InstrumentedTestAsLauncher#launch'.format(PocoServicePackage),
    #                 '{}.test/android.support.test.runner.AndroidJUnitRunner'.format(PocoServicePackage)],
    #                 not_wait=True)
    #             stdout, stderr = proc.communicate()
    #             print(stdout)
    #             print(stderr)
    #             time.sleep(1)
    #     t = threading.Thread(target=loop)
    #     t.daemon = True
    #     t.start()

    def _start_instrument(self, port_to_ping, force_restart=False):
        if not force_restart:
            try:
                state = requests.get('http://{}:{}/uiautomation/connectionState'.format(self.device_ip, port_to_ping),
                                     timeout=10)
                state = state.json()
                if state.get('connected'):
                    # skip starting instrumentation if UiAutomation Service already connected.
                    return True
            except:
                pass

        if self._instrument_proc is not None:
            self._instrument_proc.kill()
            self._instrument_proc = None

        ready = False
        self.adb_client.shell(['am', 'force-stop', PocoServicePackage])

        # 启动instrument之前，先把主类activity启动起来，不然instrumentation可能失败
        self.adb_client.shell('am start -n {}/.TestActivity'.format(PocoServicePackage))

        instrumentation_cmd = [
                'am', 'instrument', '-w', '-e', 'debug', 'false', '-e', 'class',
                '{}.InstrumentedTestAsLauncher'.format(PocoServicePackage),
                '{}.test/android.support.test.runner.AndroidJUnitRunner'.format(PocoServicePackage)]
        if new_airtest_api:
            self._instrument_proc = self.adb_client.start_shell(instrumentation_cmd)
        else:
            self._instrument_proc = self.adb_client.shell(instrumentation_cmd, not_wait=True)
        atexit.register(self._instrument_proc.kill)
        time.sleep(2)
        for i in range(10):
            try:
                requests.get('http://{}:{}'.format(self.device_ip, port_to_ping), timeout=10)
                ready = True
                break
            except requests.exceptions.Timeout:
                break
            except requests.exceptions.ConnectionError:
                if self._instrument_proc.poll() is not None:
                    warnings.warn("instrument server process is not alive")
                    output = self._instrument_proc.stdout.read()
                    print(output)
                time.sleep(1)
                print("still waiting for uiautomation ready.")
                continue
        return ready

    def on_pre_action(self, action, ui, args):
        if self.screenshot_each_action:
            # airteset log用
            from airtest.core.api import snapshot
            msg = repr(ui)
            if not isinstance(msg, six.text_type):
                msg = msg.decode('utf-8')
            snapshot(msg=msg)


class AndroidUiautomationHelper(object):
    _nuis = {}

    @classmethod
    def get_instance(cls, device):
        """
        This is only a slot to store and get already initialized poco instance rather than initializing again. You can
        simply pass the ``current device instance`` provided by ``airtest`` to get the AndroidUiautomationPoco instance.
        If no such AndroidUiautomationPoco instance, a new instance will be created and stored. 

        Args:
            device (:py:obj:`airtest.core.device.Device`): more details refer to ``airtest doc``

        Returns:
            poco instance
        """

        if cls._nuis.get(device) is None:
            cls._nuis[device] = AndroidUiautomationPoco(device)
        return cls._nuis[device]



# coding=utf-8

from poco.sdk.interfaces.screen import ScreenInterface
from poco.utils.simplerpc.utils import sync_wrapper


class StdScreen(ScreenInterface):
    def __init__(self, client):
        super(StdScreen, self).__init__()
        self.client = client

    @sync_wrapper
    def getScreen(self, width):
        return self.client.call("Screenshot", width)

    @sync_wrapper
    def getPortSize(self):
        return self.client.call("GetScreenSize")
